@@ -46,6 +46,25 @@ use proptest::prelude::*;
 use smallvec::{smallvec, SmallVec};
 
 /// Ranges represents multiple intervals of a continuous range of monotone increasing values.
+///
+/// Internally, [`Ranges`] are an ordered list of segments, where segment is a bounds pair.
+///
+/// Invariants:
+/// 1. The segments are sorted, from lowest to highest (through `Ord`).
+/// 2. Each segment contains at least one version (start < end).
+/// 3. There is at least one version between two segments.
+///
+/// These ensure that equivalent instances have an identical representation, which is important
+/// for `Eq` and `Hash`. Note that this representation cannot strictly guaranty equality of
+/// [`Ranges`] with equality of its representation without also knowing the nature of the underlying
+/// versions. In particular, if the version space is discrete, different representations, using
+/// different types of bounds (exclusive/inclusive) may correspond to the same set of existing
+/// versions. It is a tradeoff we acknowledge, but which makes representations of continuous version
+/// sets more accessible, to better handle features like pre-releases and other types of version
+/// modifiers. For example, `[(Included(3u32), Excluded(7u32))]` and
+/// `[(Included(3u32), Included(6u32))]` refer to the same version set, since there is no version
+/// between 6 and 7, which this crate doesn't know about.
+
 #[derive(Debug, Clone, Eq, PartialEq, Hash)]
 #[cfg_attr(feature = "serde", derive(serde::Serialize))]
 #[cfg_attr(feature = "serde", serde(transparent))]
@@ -210,7 +229,11 @@ impl<V: Ord> Ranges<V> {
     }
 
     /// Returns true if self contains the specified value.
-    pub fn contains(&self, version: &V) -> bool {
+    pub fn contains<Q>(&self, version: &Q) -> bool
+    where
+        V: Borrow<Q>,
+        Q: ?Sized + PartialOrd,
+    {
         self.segments
             .binary_search_by(|segment| {
                 // We have to reverse because we need the segment wrt to the version, while
@@ -283,6 +306,7 @@ impl<V: Ord> Ranges<V> {
         }
     }
 
+    /// See [`Ranges`] for the invariants checked.
     fn check_invariants(self) -> Self {
         if cfg!(debug_assertions) {
             for p in self.segments.as_slice().windows(2) {
@@ -450,10 +474,14 @@ impl<V: Ord> Ord for Ranges<V> {
 ///   ^      ^      ^
 ///   less   equal  greater
 /// ```
-fn within_bounds<V: PartialOrd>(version: &V, segment: &Interval<V>) -> Ordering {
+fn within_bounds<Q, V>(version: &Q, segment: &Interval<V>) -> Ordering
+where
+    V: Borrow<Q>,
+    Q: ?Sized + PartialOrd,
+{
     let below_lower_bound = match segment {
-        (Excluded(start), _) => version <= start,
-        (Included(start), _) => version < start,
+        (Excluded(start), _) => version <= start.borrow(),
+        (Included(start), _) => version < start.borrow(),
         (Unbounded, _) => false,
     };
     if below_lower_bound {
@@ -461,8 +489,8 @@ fn within_bounds<V: PartialOrd>(version: &V, segment: &Interval<V>) -> Ordering
     }
     let below_upper_bound = match segment {
         (_, Unbounded) => true,
-        (_, Included(end)) => version <= end,
-        (_, Excluded(end)) => version < end,
+        (_, Included(end)) => version <= end.borrow(),
+        (_, Excluded(end)) => version < end.borrow(),
     };
     if below_upper_bound {
         return Ordering::Equal;
@@ -826,6 +854,219 @@ impl<V: Ord + Clone> Ranges<V> {
     }
 }
 
+// Newtype to avoid leaking our internal representation.
+pub struct RangesIter<V>(smallvec::IntoIter<[Interval<V>; 1]>);
+
+impl<V> Iterator for RangesIter<V> {
+    type Item = Interval<V>;
+
+    fn next(&mut self) -> Option<Self::Item> {
+        self.0.next()
+    }
+
+    fn size_hint(&self) -> (usize, Option<usize>) {
+        (self.0.len(), Some(self.0.len()))
+    }
+}
+
+impl<V> ExactSizeIterator for RangesIter<V> {}
+
+impl<V> DoubleEndedIterator for RangesIter<V> {
+    fn next_back(&mut self) -> Option<Self::Item> {
+        self.0.next_back()
+    }
+}
+
+impl<V> IntoIterator for Ranges<V> {
+    type Item = (Bound<V>, Bound<V>);
+    // Newtype to avoid leaking our internal representation.
+    type IntoIter = RangesIter<V>;
+
+    fn into_iter(self) -> Self::IntoIter {
+        RangesIter(self.segments.into_iter())
+    }
+}
+
+impl<V: Ord> FromIterator<(Bound<V>, Bound<V>)> for Ranges<V> {
+    /// Constructor from arbitrary, unsorted and potentially overlapping ranges.
+    ///
+    /// This is equivalent, but faster, to computing the [`Ranges::union`] of the
+    /// [`Ranges::from_range_bounds`] of each segment.
+    fn from_iter<T: IntoIterator<Item = (Bound<V>, Bound<V>)>>(iter: T) -> Self {
+        // We have three constraints we need to fulfil:
+        // 1. The segments are sorted, from lowest to highest (through `Ord`): By sorting.
+        // 2. Each segment contains at least one version (start < end): By skipping invalid
+        //    segments.
+        // 3. There is at least one version between two segments: By merging overlapping elements.
+        //
+        // Technically, the implementation has a O(n²) worst case complexity since we're inserting
+        // and removing. This has two motivations: One is that we don't have any performance
+        // critical usages of this method as of this writing, so we have no real world benchmark.
+        // The other is that we get the elements from an iterator, so to avoid moving elements
+        // around we would first need to build a different, sorted collection with extra
+        // allocation(s), before we could build our real segments. --Konsti
+
+        // For this implementation, we choose to only build a single smallvec and insert or remove
+        // in it, instead of e.g. collecting the segments into a sorted datastructure first and then
+        // construction the second smallvec without shifting.
+        let mut segments: SmallVec<[Interval<V>; 1]> = SmallVec::new();
+
+        for segment in iter {
+            if !valid_segment(&segment.start_bound(), &segment.end_bound()) {
+                continue;
+            }
+            // Find where to insert the new segment
+            let insertion_point = segments.partition_point(|elem: &Interval<V>| {
+                cmp_bounds_start(elem.start_bound(), segment.start_bound())
+                    .unwrap()
+                    .is_lt()
+            });
+            // Is it overlapping with the previous segment?
+            let previous_overlapping = insertion_point > 0
+                && !end_before_start_with_gap(
+                    &segments[insertion_point - 1].end_bound(),
+                    &segment.start_bound(),
+                );
+
+            // Is it overlapping with the following segment? We'll check if there's more than one
+            // overlap later.
+            let next_overlapping = insertion_point < segments.len()
+                && !end_before_start_with_gap(
+                    &segment.end_bound(),
+                    &segments[insertion_point].start_bound(),
+                );
+
+            match (previous_overlapping, next_overlapping) {
+                (true, true) => {
+                    // previous:  |------|
+                    // segment:       |------|
+                    // following:          |------|
+                    // final:     |---------------|
+                    //
+                    // OR
+                    //
+                    // previous:  |------|
+                    // segment:       |-----------|
+                    // following:          |----|
+                    // final:     |---------------|
+                    //
+                    // OR
+                    //
+                    // previous:  |------|
+                    // segment:       |----------------|
+                    // following:          |----|   |------|
+                    // final:     |------------------------|
+                    // We merge all three segments into one, which is effectively removing one of
+                    // two previously inserted and changing the bounds on the other.
+
+                    // Remove all elements covered by the final element
+                    let mut following = segments.remove(insertion_point);
+                    while insertion_point < segments.len()
+                        && !end_before_start_with_gap(
+                            &segment.end_bound(),
+                            &segments[insertion_point].start_bound(),
+                        )
+                    {
+                        following = segments.remove(insertion_point);
+                    }
+
+                    // Set end to max(segment.end, <last overlapping segment>.end)
+                    if cmp_bounds_end(segment.end_bound(), following.end_bound())
+                        .unwrap()
+                        .is_lt()
+                    {
+                        segments[insertion_point - 1].1 = following.1;
+                    } else {
+                        segments[insertion_point - 1].1 = segment.1;
+                    }
+                }
+                (true, false) => {
+                    // previous:  |------|
+                    // segment:       |------|
+                    // following:                |------|
+                    //
+                    // OR
+                    //
+                    // previous:  |----------|
+                    // segment:       |---|
+                    // following:                |------|
+                    //
+                    // final:     |----------|   |------|
+                    // We can reuse the existing element by extending it.
+
+                    // Set end to max(segment.end, <previous>.end)
+                    if cmp_bounds_end(
+                        segments[insertion_point - 1].end_bound(),
+                        segment.end_bound(),
+                    )
+                    .unwrap()
+                    .is_lt()
+                    {
+                        segments[insertion_point - 1].1 = segment.1;
+                    }
+                }
+                (false, true) => {
+                    // previous:  |------|
+                    // segment:             |------|
+                    // following:               |------|
+                    // final:    |------|   |----------|
+                    //
+                    // OR
+                    //
+                    // previous:  |------|
+                    // segment:             |----------|
+                    // following:               |---|
+                    // final:    |------|   |----------|
+                    //
+                    // OR
+                    //
+                    // previous:  |------|
+                    // segment:             |------------|
+                    // following:               |---|  |------|
+                    //
+                    // final:    |------|   |-----------------|
+                    // We can reuse the existing element by extending it.
+
+                    // Remove all fully covered segments so the next element is the last one that
+                    // overlaps.
+                    while insertion_point + 1 < segments.len()
+                        && !end_before_start_with_gap(
+                            &segment.end_bound(),
+                            &segments[insertion_point + 1].start_bound(),
+                        )
+                    {
+                        // We know that the one after also overlaps, so we can drop the current
+                        // following.
+                        segments.remove(insertion_point);
+                    }
+
+                    // Set end to max(segment.end, <last overlapping segment>.end)
+                    if cmp_bounds_end(segments[insertion_point].end_bound(), segment.end_bound())
+                        .unwrap()
+                        .is_lt()
+                    {
+                        segments[insertion_point].1 = segment.1;
+                    }
+                    segments[insertion_point].0 = segment.0;
+                }
+                (false, false) => {
+                    // previous:  |------|
+                    // segment:             |------|
+                    // following:                      |------|
+                    //
+                    // final:    |------|   |------|   |------|
+
+                    // This line is O(n), which makes the algorithm O(n²), but it should be good
+                    // enough for now.
+                    segments.insert(insertion_point, segment);
+                }
+            }
+        }
+
+        Self { segments }.check_invariants()
+    }
+}
+
 // REPORT ######################################################################
 
 impl<V: Display + Eq> Display for Ranges<V> {
@@ -898,7 +1139,7 @@ impl<'de, V: serde::Deserialize<'de>> serde::Deserialize<'de> for Ranges<V> {
 pub fn proptest_strategy() -> impl Strategy<Value = Ranges<u32>> {
     (
         any::<bool>(),
-        prop::collection::vec(any::<(u32, bool)>(), 1..10),
+        prop::collection::vec(any::<(u32, bool)>(), 0..10),
     )
         .prop_map(|(start_unbounded, deltas)| {
             let mut start = if start_unbounded {
@@ -1071,7 +1312,7 @@ pub mod tests {
 
         #[test]
         fn always_contains_exact(version in version_strat()) {
-            assert!(Ranges::singleton(version).contains(&version));
+            assert!(Ranges::<u32>::singleton(version).contains(&version));
         }
 
         #[test]
@@ -1093,7 +1334,7 @@ pub mod tests {
 
         #[test]
         fn from_range_bounds(range in any::<(Bound<u32>, Bound<u32>)>(), version in version_strat()) {
-            let rv: Ranges<_> = Ranges::from_range_bounds(range);
+            let rv: Ranges<_> = Ranges::<u32>::from_range_bounds(range);
             assert_eq!(range.contains(&version), rv.contains(&version));
         }
 
@@ -1130,6 +1371,16 @@ pub mod tests {
             }
             assert!(simp.segments.len() <= range.segments.len())
         }
+
+        #[test]
+        fn from_iter_valid(segments in proptest::collection::vec(any::<(Bound<u32>, Bound<u32>)>(), ..30)) {
+            let mut expected = Ranges::empty();
+            for segment in &segments {
+                expected = expected.union(&Ranges::from_range_bounds(*segment));
+            }
+            let actual =  Ranges::from_iter(segments.clone());
+            assert_eq!(expected, actual, "{segments:?}");
+        }
     }
 
     #[test]
@@ -1150,6 +1401,17 @@ pub mod tests {
         );
     }
 
+    #[test]
+    fn contains_can_take_owned() {
+        let range: Ranges<Box<u8>> = Ranges::singleton(1);
+        let version = 1;
+
+        assert_eq!(range.contains(&Box::new(version)), range.contains(&version));
+        let range: Ranges<String> = Ranges::singleton(1.to_string());
+        let version = 1.to_string();
+        assert_eq!(range.contains(&version), range.contains("1"));
+    }
+
     #[test]
     fn simplify_can_take_owned() {
         let range: Ranges<u8> = Ranges::singleton(1);
